from haco.control import Control

from haco.tools import get_range_pair


class Text(Control):
    NAME = 'text'

    def __init__(self, name, icon=None, size_range: range = None, pattern: str = None):
        self.min, self.max = get_range_pair(size_range)
        self.pattern = pattern

        super().__init__(name, icon=icon)

    def get_config_ha_ex(self):
        data = {
            'min': self.min,
            'max': self.max,
            'pattern': self.pattern,
            'mode': self.__class__.__name__.lower()
        }

        return data


class Password(Text):
    ...



import json
from pathlib import Path
from typing import List

from haco import tools, constants
from haco.control import Control
from haco.tools import log_publish, split_into_chunks, log_subscribe


class Device:

    def __init__(self, mac, controls: List[Control], run_config_post=None):
        self.name = None
        self.topic = None
        self.hostname = None
        self.mac = mac
        self.run_config_post = run_config_post
        self.controls = controls
        self.config_id = None
        for control in self.controls:
            control.device = self

        self.contols_dict = {control.name_sanitized: control for control in self.controls}
        self.get = self.contols_dict.get

    @property
    def name_sanitized(self):
        return tools.sanitize_name(self.name)

    @property
    def mac_short(self):
        return self.mac.replace(':', '').lower()

    def get_config_ha(self):

        data_device = {
            "connections": [["mac", self.mac]],
            "identifiers": self.mac
        }

        data = {}
        for control in self.controls:
            data_control = control.get_config_ha()
            data_control['device'] = data_device
            data[control.get_announce_topic()] = data_control

        return data

    def get_config_tasmota(self):
        config = []
        for control in self.controls:
            config += control.get_config_tasmota()
        return config

    def get_subscriptions(self):
        data = {}
        for control in self.controls:
            data.update(control.get_subscriptions())

        return data

    async def publish_config_ha(self, client):
        data = self.get_config_ha()
        for topic, data in data.items():
            await log_publish(client, topic, payload=json.dumps(data), retain=True)

    async def publish_config_tasmota(self, client):
        data = self.get_config_tasmota()

        data_mqtt = {datum['trigger']: datum for datum in data if datum['type'] == 'MqttSubscription'}
        data_callbacks = [datum for datum in data if datum['type'] != 'MqttSubscription']

        tasmota_config = {
            # 'id': self.config_id,
            'callbacks': data_callbacks,
            'mqtt': data_mqtt,
            'run_config_post': self.run_config_post,

        }

        self.config_id = tools.hash_string(json.dumps(tasmota_config))

        tasmota_config['id'] = self.config_id

        tasmota_config_str = json.dumps(tasmota_config)
        tasmota_config_topic = Path(f"haco/{constants.BRANCH}/{self.hostname}/config")
        parts = split_into_chunks(tasmota_config_str, 750)
        for i, part in enumerate(parts):
            part = json.dumps({"id": self.config_id, "data": part})
            if len(part) > 1200:
                raise ValueError(f'Config part too large.')
            await log_publish(client, str(tasmota_config_topic / str(i)), payload=part, retain=True)

        payload_header = json.dumps({"size": len(parts), "id": self.config_id, 'force_update': True})
        await log_publish(client, str(tasmota_config_topic), payload=payload_header, retain=True)

        payload_full = json.dumps({"size": len(parts), "id": self.config_id, "data": tasmota_config})
        await log_publish(client, str(tasmota_config_topic / 'full'), payload=payload_full, retain=True)

    async def do_subscriptions(self, client):
        data = self.get_subscriptions()
        for topic in data.keys():
            await log_subscribe(client, str(topic))

    async def bind(self, client, data_announce):
        print(f'Found match for self: {self}')
        self.name = data_announce['device_name']
        self.topic = data_announce['topic']
        self.hostname = data_announce['hostname']

        await self.do_subscriptions(client)

        await self.publish_config_ha(client)
        await self.publish_config_tasmota(client)



from dataclasses import dataclass
from typing import ClassVar


@dataclass
class Function:
    mask: ClassVar[str] = '{code}'
    code: str

    def __str__(self):
        raw = self.mask.format(code=self.code)
        lines = [line.strip() for line in raw.splitlines()]
        line = ' '.join(lines).strip()
        return line


@dataclass
class Expression(Function):
    mask: ClassVar[str] = 'def (value,data) return {code} end'


@dataclass
class FunctionExpression(Function):

    def __str__(self):
        return f"{super().__str__()}(value,data)"



import logging
from functools import partial

from haco import schema, tools
from haco.callback import Callback
from haco.schema import Schema, Platform, Tasmota, HomeAssistant, AnnounceTopic, Capability
from haco.tools import log_publish


class Control:
    NAME = None

    control_type = str

    @classmethod
    def get_schema(cls):
        schema = Schema(
            capabilities=[
                Capability(
                    name='state',
                    ha=HomeAssistant(announce_data=AnnounceTopic(key='{io_ha}_topic')),
                    tamota=Tasmota(announce_data=AnnounceTopic(key='{io_ha}_topic'))
                )
            ]
        )
        return schema

    def __init__(self, name, icon=None):

        from device import Device

        self.name = name

        self.device: Device = None
        self.schema = self.get_schema()
        self.get = self.schema.get
        self.schema.control = self

        self.icon = icon

    @property
    def name_sanitized(self):
        return tools.sanitize_name(self.name)

    def callback(self, trigger=None):
        return partial(self.add_callback, trigger)

    def add_callback(self, trigger, function):

        if (function_name := function.__name__) in {Tasmota.PLATFORM, HomeAssistant.PLATFORM}:
            function_name = f'state_{function_name}'

        callback_names = self.schema.get_callback_function_names()

        if function_name not in callback_names:
            msg = f'Callback function "{function_name}" name is not valid. Must be one of: {sorted(callback_names)}'
            raise ValueError(msg)

        capability_id, platform_id = function_name.rsplit('_', 1)
        callback = Callback(capability=capability_id, platform=platform_id, trigger=trigger, function=function)
        platform: Platform = self.schema.get_alias(capability_id).get(platform_id)
        platform.callback = callback

    def get_subscriptions(self):
        return self.schema.get_subscriptions()

    def get_config_tasmota(self):
        return self.schema.get_config_tasmota()

    def get_config_ha_base(self):
        data = {
            "unique_id": f"{self.device.name_sanitized}-{self.device.mac_short}-{self.name_sanitized}",
            "name": self.name,
            "availability_topic": f"tele/{self.device.topic}/LWT",
            "payload_off": schema.OFF,
            "payload_available": "Online",
            "payload_on": schema.ON,
            "payload_not_available": "Offline",
            "force_update": True,
            "icon": self.icon,
            # "entity_id": self.entity_id
        }
        data.update(self.schema.get_config_ha())

        return data

    def get_config_ha(self):
        data = self.get_config_ha_base() | self.get_config_ha_ex()
        data = {k: v for k, v in data.items() if v is not None}
        return data

    def get_config_ha_ex(self):
        return {}

    async def process_message(self, client, capability_id, platform_id, message):

        if not message.payload:
            logging.warning(f'Topic {message.topic.value} got empty payload. It may have been deleted.')
            return

        platform = self.schema.get(capability_id).get(platform_id)
        output = platform.handle(message)
        topic = str(platform.get_topic_publish())
        await log_publish(client, topic, output)

    def get_announce_topic(self):
        return f"homeassistant/{self.NAME}/{self.device.name_sanitized}/{self.name_sanitized}/config"
